import sys
import importlib
import binascii


#  Constants
PUSH0 = b'\x00'  # An empty array of bytes is pushed onto the stack.
PUSHF = PUSH0
PUSHBYTES1 = b'\x01'  # b'\x01-b'\x4B The next opcode bytes is data to be pushed onto the stack
PUSHBYTES75 = b'\x4B'
PUSHDATA1 = b'\x4C'  # The next byte contains the number of bytes to be pushed onto the stack.
PUSHDATA2 = b'\x4D'  # The next two bytes contain the number of bytes to be pushed onto the stack.
PUSHDATA4 = b'\x4E'  # The next four bytes contain the number of bytes to be pushed onto the stack.
PUSHM1 = b'\x4F'  # The number -1 is pushed onto the stack.
PUSH1 = b'\x51'  # The number 1 is pushed onto the stack.
PUSHT = PUSH1
PUSH2 = b'\x52'  # The number 2 is pushed onto the stack.
PUSH3 = b'\x53'  # The number 3 is pushed onto the stack.
PUSH4 = b'\x54'  # The number 4 is pushed onto the stack.
PUSH5 = b'\x55'  # The number 5 is pushed onto the stack.
PUSH6 = b'\x56'  # The number 6 is pushed onto the stack.
PUSH7 = b'\x57'  # The number 7 is pushed onto the stack.
PUSH8 = b'\x58'  # The number 8 is pushed onto the stack.
PUSH9 = b'\x59'  # The number 9 is pushed onto the stack.
PUSH10 = b'\x5A'  # The number 10 is pushed onto the stack.
PUSH11 = b'\x5B'  # The number 11 is pushed onto the stack.
PUSH12 = b'\x5C'  # The number 12 is pushed onto the stack.
PUSH13 = b'\x5D'  # The number 13 is pushed onto the stack.
PUSH14 = b'\x5E'  # The number 14 is pushed onto the stack.
PUSH15 = b'\x5F'  # The number 15 is pushed onto the stack.
PUSH16 = b'\x60'  # The number 16 is pushed onto the stack.


#  Flow control
NOP = b'\x61'  # Does nothing.
JMP = b'\x62'
JMPIF = b'\x63'
JMPIFNOT = b'\x64'
CALL = b'\x65'
RET = b'\x66'
APPCALL = b'\x67'
SYSCALL = b'\x68'
TAILCALL = b'\x69'


#  Stack
DUPFROMALTSTACK = b'\x6A'
TOALTSTACK = b'\x6B'  # Puts the input onto the top of the alt stack. Removes it from the main stack.
FROMALTSTACK = b'\x6C'  # Puts the input onto the top of the main stack. Removes it from the alt stack.
XDROP = b'\x6D'
XSWAP = b'\x72'
XTUCK = b'\x73'
DEPTH = b'\x74'  # Puts the number of stack items onto the stack.
DROP = b'\x75'  # Removes the top stack item.
DUP = b'\x76'  # Duplicates the top stack item.
NIP = b'\x77'  # Removes the second-to-top stack item.
OVER = b'\x78'  # Copies the second-to-top stack item to the top.
PICK = b'\x79'  # The item n back in the stack is copied to the top.
ROLL = b'\x7A'  # The item n back in the stack is moved to the top.
ROT = b'\x7B'  # The top three items on the stack are rotated to the left.
SWAP = b'\x7C'  # The top two items on the stack are swapped.
TUCK = b'\x7D'  # The item at the top of the stack is copied and inserted before the second-to-top item.


#  Splice
CAT = b'\x7E'  # Concatenates two strings.
SUBSTR = b'\x7F'  # Returns a section of a string.
LEFT = b'\x80'  # Keeps only characters left of the specified point in a string.
RIGHT = b'\x81'  # Keeps only characters right of the specified point in a string.
SIZE = b'\x82'  # Returns the length of the input string.


#  Bitwise logic
INVERT = b'\x83'  # Flips all of the bits in the input.
AND = b'\x84'  # Boolean and between each bit in the inputs.
OR = b'\x85'  # Boolean or between each bit in the inputs.
XOR = b'\x86'  # Boolean exclusive or between each bit in the inputs.
EQUAL = b'\x87'  # Returns 1 if the inputs are exactly equal' 0 otherwise.
# OP_EQUALVERIFY = b'\x88' #  Same as OP_EQUAL' but runs OP_VERIFY afterward.
# OP_RESERVED1 = b'\x89' #  Transaction is invalid unless occuring in an unexecuted OP_IF branch
# OP_RESERVED2 = b'\x8A' #  Transaction is invalid unless occuring in an unexecuted OP_IF branch

#  Arithmetic
#  Note: Arithmetic inputs are limited to signed 32-bit integers' but may overflow their output.
INC = b'\x8B'  # 1 is added to the input.
DEC = b'\x8C'  # 1 is subtracted from the input.
SIGN = b'\x8D'
NEGATE = b'\x8F'  # The sign of the input is flipped.
ABS = b'\x90'  # The input is made positive.
NOT = b'\x91'  # If the input is 0 or 1' it is flipped. Otherwise the output will be 0.
NZ = b'\x92'  # Returns 0 if the input is 0. 1 otherwise.
ADD = b'\x93'  # a is added to b.
SUB = b'\x94'  # b is subtracted from a.
MUL = b'\x95'  # a is multiplied by b.
DIV = b'\x96'  # a is divided by b.
MOD = b'\x97'  # Returns the remainder after dividing a by b.
SHL = b'\x98'  # Shifts a left b bits' preserving sign.
SHR = b'\x99'  # Shifts a right b bits' preserving sign.
BOOLAND = b'\x9A'  # If both a and b are not 0' the output is 1. Otherwise 0.
BOOLOR = b'\x9B'  # If a or b is not 0' the output is 1. Otherwise 0.
NUMEQUAL = b'\x9C'  # Returns 1 if the numbers are equal' 0 otherwise.
NUMNOTEQUAL = b'\x9E'  # Returns 1 if the numbers are not equal' 0 otherwise.
LT = b'\x9F'  # Returns 1 if a is less than b' 0 otherwise.
GT = b'\xA0'  # Returns 1 if a is greater than b' 0 otherwise.
LTE = b'\xA1'  # Returns 1 if a is less than or equal to b' 0 otherwise.
GTE = b'\xA2'  # Returns 1 if a is greater than or equal to b' 0 otherwise.
MIN = b'\xA3'  # Returns the smaller of a and b.
MAX = b'\xA4'  # Returns the larger of a and b.
WITHIN = b'\xA5'  # Returns 1 if x is within the specified range (left-inclusive)' 0 otherwise.

#  Crypto
# RIPEMD160 = b'\xA6' #  The input is hashed using RIPEMD-160.
SHA1 = b'\xA7'  # The input is hashed using SHA-1.
SHA256 = b'\xA8'  # The input is hashed using SHA-256.
HASH160 = b'\xA9'
HASH256 = b'\xAA'
CHECKSIG = b'\xAC'
CHECKMULTISIG = b'\xAE'


#  Array
ARRAYSIZE = b'\xC0'
PACK = b'\xC1'
UNPACK = b'\xC2'
PICKITEM = b'\xC3'
SETITEM = b'\xC4'
NEWARRAY = b'\xC5'  # 用作引用類型
NEWSTRUCT = b'\xC6'  # 用作值類型

# the following is a convienience method
# for a human readable version of the ops

module = importlib.import_module('boa.blockchain.vm.VMOp')
items = dir(sys.modules[__name__])


def ToName(op):

    """

    :param op:
    :return:
    """
    if isinstance(op, bytes):
        op = int.from_bytes(op, 'little')

    for item in items:
        n = getattr(module, item)

        try:
            nn = int(binascii.hexlify(n))

            if op == nn:
                return item
        except Exception as e:
            pass

        try:
            nn2 = int.from_bytes(n, 'little')
            if op == nn2:
                return item
        except Exception as e:
            pass

    return None




def Main(a, b, c, d):

    """

    :param a:
    :param b:
    :param c:
    :param d:
    :return:
    """
    j = a & b

    e = c / d

    f = a - c

    g = f * j

    k = e % d

    l = g ^ c

    m = j + e + f + g + l

    return m




class TransactionOutput():

    @property
    def AssetId(self):
        """

        :return:
        """
        return GetAssetId(self)

    @property
    def Value(self):
        """

        :return:
        """
        return GetValue(self)

    @property
    def ScriptHash(self):
        """

        :return:
        """
        return GetScriptHash(self)


def GetAssetId(output):
    """

    :param output:
    """
    pass


def GetValue(output):
    """

    :param output:
    """
    pass


def GetScriptHash(output):
    """

    :param output:
    """
    pass





def GetContext():
    """

    """
    pass


# def CurrentContext():
#    return GetContext()

def Get(context, key):
    """

    :param context:
    :param key:
    """
    pass


def Put(context, key, value):
    """

    :param context:
    :param key:
    :param value:
    """
    pass


def Delete(context, key):
    """

    :param context:
    :param key:
    """
    pass




from boa.code.builtins import list
from boa.blockchain.vm.Neo.Runtime import Notify


def Main():

    """

    :return:
    """
    start = 4
    stop = 9  # int

#    out = [10,2, 3, 4, 6, 7]

    length = stop - start

    out = list(length=length)

    index = 0
    orig_start = start

    while start < stop:
        val = index + orig_start
        out[index] = val
        index = index + 1
        start = orig_start + index

        Notify(start)

        # d = stuff(1, 2) # this doesn't work at the moment

    return out[4]


def stuff(a, b):

    """

    :param a:
    :param b:
    :return:
    """
    out = a + b
    return out





class Header():

    @property
    def Hash(self):
        """

        :return:
        """
        return GetHash(self)

    @property
    def Timestamp(self):
        """

        :return:
        """
        return GetTimestamp(self)

    @property
    def Version(self):
        """

        :return:
        """
        return GetVersion(self)

    @property
    def PrevHash(self):
        """

        :return:
        """
        return GetPrevHash(self)

    @property
    def MerkleRoot(self):
        """

        :return:
        """
        return GetMerkleRoot(self)

    @property
    def ConsensusData(self):
        """

        :return:
        """
        return GetConsensusData(self)

    @property
    def NextConsensus(self):
        """

        :return:
        """
        return GetNextConsensus(self)


def GetHash(header: Header) -> bytearray:
    """

    gets the hash of the header

    """
    pass


def GetVersion(header: Header) -> int:
    """

    gets the version of the header

    """
    pass


def GetPrevHash(header: Header) -> bytearray:
    """

    gets the hash of the previous header in the blockchain

    """
    pass


def GetMerkleRoot(header: Header) -> bytearray:
    """

    gets the merkle root of the transactions contained in the block

    """
    pass


def GetTimestamp(header: Header) -> int:
    """

    gets the timestamp of when the header was created

    """
    pass


def GetConsensusData(header: Header) -> bytearray:
    """

    gets the address of the consensus

    """
    pass


def GetNextConsensus(header: Header) -> bytearray:
    """

    gets the address where the next consensus will occur

    """
    pass



"""
NEP5 Token
===================================

This file, when compiled to .avm format, would comply with the current NEP5 token standard on the NEO blockchain

Token standard is available in proposal form here:
`NEP5 Token Standard Proposal <https://github.com/neo-project/proposals/blob/master/nep-5.mediawiki>`_

Compilation can be achieved as such

>>> from boa.compiler import Compiler
>>> Compiler.load_and_save('./boa/tests/src/NEP5Test.py')


Below is the current implementation in Python


"""

from boa.blockchain.vm.Neo.Runtime import Log, Notify
from boa.blockchain.vm.System.ExecutionEngine import GetScriptContainer, GetExecutingScriptHash
from boa.blockchain.vm.Neo.Transaction import *
from boa.blockchain.vm.Neo.Blockchain import GetHeight, GetHeader
from boa.blockchain.vm.Neo.Action import RegisterAction
from boa.blockchain.vm.Neo.Runtime import GetTrigger, CheckWitness
from boa.blockchain.vm.Neo.TriggerType import Application, Verification
from boa.blockchain.vm.Neo.Output import GetScriptHash, GetValue, GetAssetId
from boa.blockchain.vm.Neo.Storage import GetContext, Get, Put, Delete


# -------------------------------------------
# TOKEN SETTINGS
# -------------------------------------------

TOKEN_NAME = 'LOCALTOKEN'
SYMBOL = 'LWTF'

OWNER = b'F\xc2\xbb\x9c\x17Ci\x89\xca\xa7\x85>|\xbd\x87B>H#\xf2'

DECIMALS = 8

FACTOR = 100000000


# -------------------------------------------
# ICO SETTINGS
# -------------------------------------------

NEO_ASSET_ID = b'\x9b|\xff\xda\xa6t\xbe\xae\x0f\x93\x0e\xbe`\x85\xaf\x90\x93\xe5\xfeV\xb3J\\"\x0c\xcd\xcfn\xfc3o\xc5'

# 5million times decimals ( factor )
TOTAL_AMOUNT = 500000000000000

PRE_ICO_CAP = 1000000000  # amount for the owner to start with

ICO_START_TIME = 1502726400  # August 14 2017
ICO_END_TIME = 1513936000  # December 22 2017

# -------------------------------------------
# Events
# -------------------------------------------

OnTransfer = RegisterAction('transfer', 'from', 'to', 'amount')

OnRefund = RegisterAction('refund', 'to', 'amount')


def Main(operation, args):

    """
    This is the main entry point for the Smart Contract

    :param operation: the operation to be performed ( eg `mintTokens`, `transfer`, etc)
    :type operation: str

    :param args: an optional list of arguments
    :type args: list

    :return: indicating the successful execution of the smart contract
    :rtype: bool
    """

    trigger = GetTrigger()

    if trigger == Verification():

        print("doing verification!")
        owner_len = len(OWNER)

        if owner_len == 20:

            res = CheckWitness(OWNER)
            print("owner verify result")
            return res

#        elif owner_len == 33:
#            #res = verify_signature(operation, OWNER)
#            Log("verify signature not understood by me yet")

    elif trigger == Application():

        print("doing application!")

        if operation == 'deploy':
            out = Deploy()
            print("deployed!")
            return out
        elif operation == 'mintTokens':
            domint = MintTokens()
            print("minted token!")
            return domint
        elif operation == 'totalSupply':
            supply = TotalSupply()
            print("got total supply")
            Notify(supply)
            return supply
        elif operation == 'name':
            n = Name()
            return n
        elif operation == 'decimals':
            d = Decimals()
            return d
        elif operation == 'symbol':
            sym = Symbol()
            return sym

        elif operation == 'transfer':
            print("do transfers")
            if len(args) == 3:
                t_from = args[0]
                t_to = args[1]
                t_amount = args[2]
                return DoTransfer(t_from, t_to, t_amount)

            else:
                return False

        elif operation == 'balanceOf':
            if len(args) == 1:

                print("do balance")

                account = args[0]

                balance = BalanceOf(account)
                print("got balance")
                Notify(balance)

                return balance

            else:

                return 0

    return False


def Name():
    """
    Method that returns the name of this NEP5 token

    :return: name of the token
    :rtype: str

    """
    print("getting name!")
    return TOKEN_NAME


def Symbol():
    """
    Method that returns the symbol of this NEP5 token

    :return: symbol of the token
    :rtype: str

    """
    print("getting symbol!")
    return SYMBOL


def Decimals():
    """
    Method that returns the number of decimals an NEP5 token uses

    :return: the number of decimals this NEP5 token uses
    :rtype: int

    """
    print("getting decimals...")
    return DECIMALS


def Deploy():
    """
    Method for the NEP5 Token owner to use in order to deploy an initial amount of tokens to their own address

    :return: whether the deploy was successful
    :rtype: bool
    """
    print("deploying!")

    isowner = CheckWitness(OWNER)

    if isowner:

        print("ok to deploy")
        context = GetContext()

        total = Get(context, 'totalSupply')

        if len(total) == 0:

            Log("WILL DEPLOY!")

            Put(context, OWNER, PRE_ICO_CAP)

            Put(context, "totalSupply", PRE_ICO_CAP)

            OnTransfer(0, OWNER, PRE_ICO_CAP)

            return True
        else:
            print("ALREADY DEPLOYED, wont do it again")

    print("only owner can deploy")
    return False


def MintTokens():
    """
    Method for an address to call in order to deposit NEO into the NEP5 token owner's address in exchange for a calculated amount of NEP5 tokens

    :return: whether the token minting was successful
    :rtype: bool

    """
    print("minting tokens!")

    tx = GetScriptContainer()

    references = tx.References

    print("helol1")
    if len(references) < 1:
        print("no neo attached")
        return False

    print("hello2")
    reference = references[0]
    print("hello2")
#    sender = reference.ScriptHash

    sender = GetScriptHash(reference)
    print("hello4")

    value = 0
    print("hello5")
    output_asset_id = GetAssetId(reference)
    if output_asset_id == NEO_ASSET_ID:

        print("hello6")
        receiver = GetExecutingScriptHash()
        print("hello7")
        for output in tx.Outputs:
            shash = GetScriptHash(output)
            print("getting shash..")
            if shash == receiver:
                print("adding value?")
                output_val = GetValue(output)
                value = value + output_val

        print("getting rate")
        rate = CurrentSwapRate()
        print("got rate")
        if rate == 0:
            OnRefund(sender, value)
            return False

        num_tokens = value * rate / 100000000

        context = GetContext()

        balance = Get(context, sender)

        new_total = num_tokens + balance

        Put(context, sender, new_total)

        total_supply = Get(context, 'totalSupply')

        new_total_supply = total_supply + num_tokens

        Put(context, 'totalSupply', new_total_supply)

        OnTransfer(0, sender, num_tokens)

        return True

    return False


def TotalSupply():

    """
    Method to return the total amount of NEP5 tokens in current circluation

    :return: the total number of tokens in circulation
    :rtype: int

    """
    print("total supply!")

    context = GetContext()

    res = Get(context, "totalSupply")

    print("got total supply")
    Notify(res)

    return res


def DoTransfer(t_from, t_to, amount):

    """
    Method to transfer NEP5 tokens of a specified amount from one account to another

    :param t_from: the address to transfer from
    :type t_from: bytearray

    :param t_to: the address to transfer to
    :type t_to: bytearray

    :param amount: the amount of NEP5 tokens to transfer
    :type amount: int

    :return: whether the transfer was successful
    :rtype: bool

    """
    if amount <= 0:
        print("cannot transfer zero or less")
        return False

    from_is_sender = CheckWitness(t_from)

    if from_is_sender:

        if t_from == t_to:
            return True

        context = GetContext()

        from_val = Get(context, t_from)

        if from_val < amount:
            print("Insufficient funds")
            return False

        if from_val == amount:
            print("Removing all funds!")
            Delete(context, t_from)

        else:
            difference = from_val - amount
            Put(context, t_from, difference)

        to_value = Get(context, t_to)

        to_total = to_value + amount

        Put(context, t_to, to_total)

        OnTransfer(t_from, t_to, amount)

        return True
    else:
        print("from address is not the tx sender")

    return False


def BalanceOf(account):

    """
    Method to return the current balance of an address

    :param account: the account address to retrieve the balance for
    :type account: bytearray

    :return: the current balance of an address
    :rtype: int

    """
    print("getting balance of...")
    context = GetContext()
    print("getting context...")
    balance = Get(context, account)
    print("got balance...")

    return balance


def CurrentSwapRate():

    """
    Method to calculate the current 'going rate' or exchange ratio of NEO to NEP5 tokens

    :return: the current rate
    :rtype: int

    """
    basic = 1000 * FACTOR
    duration = ICO_END_TIME - ICO_START_TIME
    print("getting swap rate")
    context = GetContext()
    print("got context")

    total_supply = Get(context, 'totalSupply')
    print("got total supply")
    if total_supply >= TOTAL_AMOUNT:
        return False
    print("getting current height...")
    currentHeight = GetHeight()
    print("got current height")
    currentBlock = GetHeader(currentHeight)
    print("got current block...")
    time = currentBlock.Timestamp - ICO_START_TIME

    if time < 0:

        return 0

    elif time < 86400:
        return basic * 130 / 100

    elif time < 259200:
        return basic * 120 / 100

    elif time < 604800:
        return basic * 110 / 100

    elif time < duration:
        return basic

    return 0




class ExecutionEngine():
    """
    Not used.

    - Note: This method is implemented inside the Neo Virtual Machine.
    """
    
    pass


def GetScriptContainer():
    """
    Return the current Script Container of a smart contract execution. 
    This will be a ``boa.blockchain.vm.Neo.Transaction`` object.

    - Note: This method is implemented inside the Neo Virtual Machine.

    :return: the current ScriptContainer of a smart contract execution.
    :rtype: ``boa.blockchain.vm.Neo.Transaction``
    """
    
    pass


def GetExecutingScriptHash():
    """
    Get the hash of the script ( smart contract ) which is currently being executed

    - Note: This method is implemented inside the Neo Virtual Machine.

    :return: the hash of the script ( smart contract ) which is currently being executed
    :rtype: bytearray
    """
    
    pass


def GetCallingScriptHash():
    """
    Get the hash of the script ( smart contract ) which began execution of the current script.

    - Note: This method is implemented inside the Neo Virtual Machine.

    :return: the hash of the script ( smart contract ) which began execution of the current script
    :rtype: bytearray
    """
    
    pass


def GetEntryScriptHash():
    """
    Get the hash of the script ( smart contract ) which began execution of the smart contract.

    - Note: This method is implemented inside the Neo Virtual Machine.

    :return: the hash of the script ( smart contract ) which began execution of the smart contract
    :rtype: bytearray
    """
    
    pass




class Contract():

    @property
    def Script(self):
        """

        :return:
        """
        return GetScript(self)

    @property
    def StorageContext(self):
        """

        :return:
        """
        return GetStorageContext(self)


def GetScript(contract):
    """

    :param contract:
    """
    pass


def GetStorageContext(contract):
    """

    :param contract:
    """
    pass


def Create(script,
           parameter_list,
           return_type,
           need_storage,
           version,
           author,
           email,
           description
           ):

    """

    :param script:
    :param parameter_list:
    :param return_type:
    :param need_storage:
    :param version:
    :param author:
    :param email:
    :param description:
    """
    pass


def Migrate(script,
            parameter_list,
            return_type,
            need_storage,
            version,
            author,
            email,
            description
            ):

    """

    :param script:
    :param parameter_list:
    :param return_type:
    :param need_storage:
    :param version:
    :param author:
    :param email:
    :param description:
    """
    pass


def Destroy(contract):
    """

    :param contract:
    """
    pass





class list(list):

    """

    """

    def __init__(self, length=0):
        #        super(list, self).__init__()
        pass

    """
    list() -> new empty list
    list(length=x) -> new list initialized with length of x
    """


#    The following items are not currently supported by the compiler or VM
#
#    def append(self, p_object):  # real signature unknown; restored from __doc__
#        """ L.append(object) -> None -- append object to end """
#        pass
#
#    def clear(self):  # real signature unknown; restored from __doc__
#        """ L.clear() -> None -- remove all items from L """
#        pass
#
#    def copy(self):  # real signature unknown; restored from __doc__
#        """ L.copy() -> list -- a shallow copy of L """
#        return []
#
#    def count(self, value):  # real signature unknown; restored from __doc__
#        """ L.count(value) -> integer -- return number of occurrences of value """
#        return 0
#
#    def extend(self, iterable):  # real signature unknown; restored from __doc__
#        """ L.extend(iterable) -> None -- extend list by appending elements from the iterable """
#        pass
#
#    def index(self, value, start=None, stop=None):  # real signature unknown; restored from __doc__
#        """
#        L.index(value, [start, [stop]]) -> integer -- return first index of value.
#        Raises ValueError if the value is not present.
#        """
#        return 0
#
#    def insert(self, index, p_object):  # real signature unknown; restored from __doc__
#        """ L.insert(index, object) -- insert object before index """
#        pass
#
#    def pop(self, index=None):  # real signature unknown; restored from __doc__
#        """
#        L.pop([index]) -> item -- remove and return item at index (default last).
#        Raises IndexError if list is empty or index is out of range.
#        """
#        pass
#
#    def remove(self, value):  # real signature unknown; restored from __doc__
#        """
#        L.remove(value) -> None -- remove first occurrence of value.
#        Raises ValueError if the value is not present.
#        """
#        pass
#
#    def reverse(self):  # real signature unknown; restored from __doc__
#        """ L.reverse() -- reverse *IN PLACE* """
#        pass
#
#    def sort(self, key=None, reverse=False):  # real signature unknown; restored from __doc__
#        """ L.sort(key=None, reverse=False) -> None -- stable sort *IN PLACE* """
#        pass

    def __contains__(self, *args, **kwargs):  # real signature unknown
        """ Return key in self. """
        pass

    def __eq__(self, *args, **kwargs):  # real signature unknown
        """ Return self==value. """
        pass

    def __getitem__(self, y):  # real signature unknown; restored from __doc__
        """ x.__getitem__(y) <==> x[y] """
        pass

    def __setitem__(self, *args, **kwargs):  # real signature unknown
        """ Set self[key] to value. """
        pass


def concat(str1, str2):
    """
     range(str1, str2) -> str object

     Return a string that is the concatenation of the two arguments ( str1 + str2 )
     """
    pass


# This is not necessary.  you can use mystring[start:end]
# Actually, it is currently necessary, as mystring[start:end] is not working as expected

def substr(source,start_index, count):
    """
    substr(source, start_index, count) -> list object

    Return a subset of a string `source`, starting at `start_index` and
    of length `count`
    """
    pass


def take(source, count):
    """
    take(source, count) -> list object

    Return a subset of a string or list `source`, starting
    at index 0 and of length `count`
    """
    pass


def range(start, stop):
    """
    range(start, stop) -> list object

    Return an list that is a a sequence of integers from start (inclusive)
    to stop (exclusive).  range(i, j) produces i, i+1, i+2, ..., j-1.
    """

    length = stop - start

    out = list(length=length)

    index = 0

    orig_start = start

    while start < stop:
        val = index + orig_start
        out[index] = val
        index = index + 1
        start = orig_start + index

    return out



def sha1(data):
    """

    :param data:
    """
    pass


def sha256(data):
    """

    :param data:
    """
    pass


def hash160(data):
    """

    :param data:
    """
    pass


def hash256(data):
    """

    :param data:
    """
    pass


def verify_signature(signature, pubkey):
    """

    :param signature:
    :param pubkey:
    """
    pass


def verify_signatures(signatures, pubkeys):
    """

    :param signatures:
    :param pubkeys:
    """
    pass
